mod function;
mod pair;
mod string;

pub use function::{Function, NativeFun};
pub use pair::DotPair;
pub use string::Str;

use crate::syntax::ast::{Expr, Spanned};
use pair::list;
use std::{collections::BTreeMap, fmt::Display, rc::Rc};

pub type OrderedF64 = ordered_float::OrderedFloat<f64>;

#[derive(Debug, Default, Clone, PartialEq, Eq)]
pub enum Value<'s> {
    #[default]
    Nil,
    DotPair(DotPair<'s>),
    Vector(Rc<Vec<Self>>),
    Map(Rc<BTreeMap<Self, Self>>),
    Symbol(Str<'s>),
    Keyword(Str<'s>),
    Number(OrderedF64),
    String(Str<'s>),
    Function(Rc<Function<'s>>),
    NativeFun(NativeFun<'s>),
    Macro(Rc<Function<'s>>),
}

impl<'s> From<Spanned<Expr<'s>>> for Value<'s> {
    fn from(e: Spanned<Expr<'s>>) -> Self {
        e.item.into()
    }
}

impl<'s> From<Expr<'s>> for Value<'s> {
    fn from(e: Expr<'s>) -> Self {
        match e {
            Expr::List(v) => v.into_iter().rev().fold(Self::Nil, |acc, x| {
                Value::DotPair(DotPair::new(x.into(), acc))
            }),
            Expr::Vector(v) => Self::Vector(Rc::new(v.into_iter().map(Into::into).collect())),
            Expr::Pair((l, r)) => Self::DotPair(DotPair::new((*l).into(), (*r).into())),
            Expr::Quote(e) => list!(Self::Symbol("quote".into()), (*e).into()),
            Expr::Symbol("nil") => Self::Nil,
            Expr::Symbol(s) => Self::Symbol(s.into()),
            Expr::Keyword(s) => Self::Keyword(s.into()),
            Expr::Number(n) => Self::Number(n),
            Expr::String(s) => Self::String(s.into()),
        }
    }
}

impl<'s> Display for Value<'s> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Value::Nil => "nil".fmt(f),
            Value::DotPair(p) => p.fmt(f),
            Value::Vector(v) => fmt_sequence(f, v.as_ref(), "#(", ")", |i, f| i.fmt(f)),
            Value::Map(m) => {
                fmt_sequence(f, m.as_ref(), "(", ")", |(k, v), f| write!(f, "{k} -> {v}"))
            }
            Value::Symbol(s) => s.fmt(f),
            Value::Keyword(k) => write!(f, ":{k}"),
            Value::Number(n) => n.fmt(f),
            Value::String(s) => write!(f, "\"{s}\""),
            Value::Function(_) => "#fun#".fmt(f),
            Value::NativeFun(fp) => write!(f, "#native-fun({fp:p})#"),
            Value::Macro(_) => "#macro#".fmt(f),
        }
    }
}

fn fmt_sequence<T>(
    f: &mut std::fmt::Formatter<'_>,
    s: impl IntoIterator<Item = T>,
    start: &str,
    end: &str,
    it_fmt: impl Fn(T, &mut std::fmt::Formatter<'_>) -> std::fmt::Result,
) -> std::fmt::Result {
    let mut it = s.into_iter();
    start.fmt(f)?;
    if let Some(i) = it.next() {
        it_fmt(i, f)?;
    }
    for i in it {
        " ".fmt(f)?;
        it_fmt(i, f)?;
    }
    end.fmt(f)
}
