use std::{fmt::Display, hash::Hash, ops::Deref, rc::Rc};

/// A Wisp string type which can hold two variants:
/// - Borrowed (usually a reference to source code or built-in literal)
/// - Shared (Ref-counted pointer to [`str`])
#[derive(Debug, Clone)]
pub enum Str<'a> {
    Borrowed(&'a str),
    Shared(Rc<str>),
}

impl<'a> From<&'a str> for Str<'a> {
    fn from(x: &'a str) -> Self {
        Self::Borrowed(x)
    }
}

impl<'a> From<Rc<str>> for Str<'a> {
    fn from(x: Rc<str>) -> Self {
        Self::Shared(x)
    }
}

impl<'a> Deref for Str<'a> {
    type Target = str;

    fn deref(&self) -> &Self::Target {
        match self {
            Self::Borrowed(b) => b,
            Self::Shared(s) => s,
        }
    }
}

impl<'a> PartialEq for Str<'a> {
    fn eq(&self, other: &Self) -> bool {
        **self == **other
    }
}

impl<'a> Eq for Str<'a> {}

impl<'a> Hash for Str<'a> {
    fn hash<H: std::hash::Hasher>(&self, state: &mut H) {
        (**self).hash(state);
    }
}

impl<'a> Display for Str<'a> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Str::Borrowed(b) => *b,
            Str::Shared(s) => s,
        }
        .fmt(f)
    }
}
