use logos::{Lexer, Logos};
use ordered_float::OrderedFloat;
use std::fmt::Display;

#[derive(Clone, Debug, Hash, Logos, PartialEq, Eq, PartialOrd, Ord)]
pub enum Token<'a> {
    // Symbols
    #[token("(")]
    LeftParen,

    #[token(")")]
    RightParen,

    #[token(".")]
    Dot,

    #[token("'")]
    Quote,

    #[token("#")]
    Octothorpe,

    // Values
    #[regex("\"(\\.|[^\"])*\"", (lex_slice::<1, 1>))]
    String(&'a str),

    #[regex("[+-]?([0-9]*[.])?[0-9]+", lex_float)]
    Number(OrderedFloat<f64>),

    // TODO: Better symbol definition
    #[regex(r"\p{XID_Start}[\p{XID_Continue}|-]*")]
    Symbol(&'a str),

    #[regex(r":\p{XID_Start}[\p{XID_Continue}|-]*", (lex_slice::<1, 0>))]
    Keyword(&'a str),

    #[regex(r";(.|[^\n])*\n", logos::skip)]
    #[regex(r"[ \t\n\f]+", logos::skip)]
    #[error]
    Error,
}

fn lex_slice<'a, const S: usize, const E: usize>(lexer: &mut Lexer<'a, Token<'a>>) -> &'a str {
    let slice = lexer.slice();
    &slice[S..slice.len() - E]
}

fn lex_float<'a>(lexer: &mut Lexer<'a, Token<'a>>) -> Option<OrderedFloat<f64>> {
    lexer.slice().parse().ok()
}

impl<'a> Display for Token<'a> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Token::LeftParen => write!(f, "("),
            Token::RightParen => write!(f, ")"),
            Token::Dot => write!(f, "."),
            Token::Quote => write!(f, "'"),
            Token::Octothorpe => write!(f, "#"),
            Token::String(s) => write!(f, "\"{s}\""),
            Token::Number(n) => write!(f, "{n}"),
            Token::Symbol(sym) => write!(f, "{sym}"),
            Token::Keyword(kw) => write!(f, ":{kw}"),
            Token::Error => write!(f, "Invalid token"),
        }
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use Token::*;

    fn assert_lex(src: &str, expected: &[Token]) {
        assert_eq!(Token::lexer(src).collect::<Vec<_>>(), expected);
    }
    #[test]
    fn gibberish() {
        assert_lex(
            "()'#\"ABLE\"corp :wisp 6.2831853071¨",
            &[
                LeftParen,
                RightParen,
                Quote,
                Octothorpe,
                String("ABLE"),
                Symbol("corp"),
                Keyword("wisp"),
                Number(6.2831853071.into()),
                Error,
            ],
        )
    }

    #[test]
    fn function() {
        assert_lex(
            "(defun print-add (x y) (println (add x y))) (print-add 42 12)",
            &[
                LeftParen,
                Symbol("defun"),
                Symbol("print-add"),
                LeftParen,
                Symbol("x"),
                Symbol("y"),
                RightParen,
                LeftParen,
                Symbol("println"),
                LeftParen,
                Symbol("add"),
                Symbol("x"),
                Symbol("y"),
                RightParen,
                RightParen,
                RightParen,
                LeftParen,
                Symbol("print-add"),
                Number(42.0.into()),
                Number(12.0.into()),
                RightParen,
            ],
        )
    }
}
